#!/bin/sh
# FILE: pimake.tcl
#
# A platform-independent make -- written entirely in Tcl, so it runs on all
# platforms to which Tcl 7.5 or later has been ported.
#
# On Unix platforms, it may be necessary to edit line 16 of this file
# so that the second word is the name of a Tcl interpreter shell 
# installed on your computer. 
#
# On Windows platforms, when Tcl/Tk was installed, an association was
# placed in the registry to recognize the file extension .tcl of this
# file and use the installed wish program to evaluate it.
#
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}
#
# Usage:
# -----
# pimake.tcl ?options? ?target?
# This whole script is wrapped in a 'catch' so we can process any errors
# in such a way that the error message is displayed to the user 
# regardless of platform and Tcl version
if {[catch {

# Process command line arguments
while {[string match -* [set word [lindex $argv 0]]]} {
    set argv [lrange $argv 1 end]
    if {[string match - [string range $word 1 end]]} {
        break
    }
    set options($word) 1
}

if {[llength $argv] > 1} {
    error "" "Usage: pimake.tcl ?options? ?target?\n\tActive options:\
            -i -k --"
}

# Robust determination of the absolute, direct pathname of this script:
set cwd [pwd]
set myAbsoluteName [file join $cwd [info script]]
if {[catch {cd [file dirname $myAbsoluteName]} msg]} {
    error "" "Can't determine direct pathname for\
            $myAbsoluteName:\n\t$msg"
}
set myAbsoluteName [file join [pwd] [file tail $myAbsoluteName]]
cd $cwd

# Create search path for OOMMF runtime library
if {[info exists env(OOMMF_LIBRARY)] \
        && [string match absolute [file pathtype $env(OOMMF_LIBRARY)]] \
        && [file isdirectory [file join $env(OOMMF_LIBRARY) app]]} {
    lappend oommf(libPath) $env(OOMMF_LIBRARY)
}
set _ [file join [file dirname [file dirname $myAbsoluteName]] \
        lib oommf]
if {[file isdirectory [file join $_ app]]} {
    lappend oommf(libPath) $_
}
set _ [file dirname [file dirname [file dirname $myAbsoluteName]]]
if {[file isdirectory [file join $_ app]]} {
    lappend oommf(libPath) $_
}
if {![llength $oommf(libPath)]} {
    error "" "Unable to find the OOMMF runtime library.\n\tSet the\
             environment variable OOMMF_LIBRARY\n\tto its absolute\
             location."
}

# Construct search path for extensions
foreach lib $oommf(libPath) {
    set dir [file join $lib ext]
    if {[file isdirectory $dir] && [file readable $dir] \
            && [string match absolute [file pathtype $dir]]} {
        if {[lsearch -exact $auto_path $dir] < 0} {
            lappend auto_path $dir
        }
    }
}

# Get support from the Oc extension
if {[catch {package require Oc}]} {
    package require Oc
}

# Define the MakeRule class 
source [file join [file dirname [info script]] makerule.tcl]
source [file join [file dirname [info script]] csourcefile.tcl]
source [file join [file dirname [info script]] platform.tcl]
#
# Define support routines
source [file join [file dirname [info script]] procs.tcl]

# Pass options to MakeRule class
MakeRule SetOptions [array get options]

# Evaluate the makerules.tcl file in the current directory
if {![MakeRule SourceRulesFile]} {
    error "" "Can't find or read makerules file 'makerules.tcl'"
}

set target [lindex $argv 0]
if {[string length $target]} {
    set target [Oc_DirectPathname $target]
}
if {[catch {MakeRule Build $target} errmsg] || [MakeRule NumErrors]} {
    set index [string last "\n    ('MakeRule' proc 'Build'" $errorInfo]
    set ei [string trimright [string range $errorInfo 0 $index]]
error "" "$ei\n[MakeRule NumErrors] Error(s) detected"
} else {
    puts "Target '$target' up to date.
Built [clock format [expr {$errmsg - [MakeRule Offset]}]]
Current time: [clock format [clock seconds]]"
    # In case this script is sourced by a 'wish' shell, explicitly call 'exit'
    # so we don't enter the event loop
    exit 0
}

}]} {
    # The wish shell didn't portably display errors in startup scripts to 
    # the user until Tk 8.0.  If an earlier version of Tk is evaluating 
    # this script, we have to display the message ourselves.  The 'tk_dialog'
    # command is most reliable since it's been around the longest.
    set index [string last "\n    invoked from within" $errorInfo]
    set ei [string trimright [string range $errorInfo 0 $index]]
    if {[info exists tk_version] && $tk_version < 8} {
        wm withdraw .
        option add *Dialog.msg.wrapLength 0
        tk_dialog .error pimake $ei error 0 OK
        exit 1
    }
}
error "" \n$ei

