# FILE: host.tcl
#
# The master server for an OOMMF host.
#
# It listens on the port name by its first command line argument, and 
# directs communications among other OOMMF threads on the localhost.
#

if {([llength $argv] > 2) || ([llength $argv] == 0)} {
    error "usage: host.tcl <service_port> ?<creator_port>?"
}

global master
set master(version) 0.0

# Set up for auto-loading
set master(relhdir) [string trimright [file dirname [info script]] ./]
set master(abshdir) [file join [pwd] $master(relhdir)]
set master(pdir) [file dirname $master(abshdir)]
set master(libdir) [file dirname $master(pdir)]

if {[lsearch $auto_path $master(libdir)] < 0} {
    lappend auto_path $master(libdir)
}

# Support libraries
package require Oc 
package require Net

# Ignore Ctrl-C's, Ctrl-Z's, and child deaths.
Oc_IgnoreInteractiveSignals

# Default timeout = 15 minutes
set master(timeout) 900000
Oc_Option Get Demon timeout master(timeout)

# Log info
set master(log) [em_extension ApplicationLog]

# After idle for timeout period after all accounts deregister, die.
proc Expire {} {
    global master
    $master(server) Delete
    $master(protocol) Delete
    $master(log) Append "OOMMF host server expired"
    exit
}

# Define protocol
Net_Protocol New master(protocol) \
        -name [list OOMMF host protocol $master(version)]
$master(protocol) AddMessage start lookup {acct} {
    global portmap
    if {[info exists portmap($acct)]} {
        return [list start [list 0 $portmap($acct)]]
    } else {
        return [list start [list 1 $acct not registered]]
    }
}
$master(protocol) AddMessage start register {acct port} {
    global portmap expire_event
    if {[info exists portmap($acct)]} {
        return [list start \
                [list 1 $acct already registered at port $portmap($acct)]]
    }
    set ret [list start [list 0 [array set portmap [list $acct $port]]]]
    if {[info exists expire_event] && [array size portmap]} {
        after cancel $expire_event
        unset expire_event
    }
    return $ret
}
$master(protocol) AddMessage start deregister { acct port } {
    # Ought to call back for security -- and that will require
    # event-driven query-handling.
    #
    # If this request comes in from a remote machine, it should not
    # be trusted without checking.
    global portmap expire_event master
    if {![info exists portmap($acct)]} {
        return [list start [list 1 $acct not registered]]
    }
    if {![string match $port $portmap($acct)]} {
        return [list start [list 1 permission denied]]
    }
    set ret [list start [list 0 [unset portmap($acct)]]]
    if {![array size portmap] && $master(timeout)} {
        set expire_event [after $master(timeout) Expire]
    }
    return $ret
}
$master(protocol) AddMessage start bye {} {
    return [list close [list 0 Bye!]]
}

# Define server and start it
Net_Server New master(server) -protocol $master(protocol) \
        -log $master(log) -register 0
set serviceport [lindex $argv 0]
if {[catch {$master(server) Start $serviceport} master(msg)]} {
    $master(log) Append "Can't start OOMMF host server on port $serviceport:
	$master(msg)."
    error "Can't start OOMMF host server on port $serviceport:
	$master(msg)."
}
if {$master(timeout)} {
    set expire_event [after $master(timeout) Expire]
}

if {[llength $argv] == 2} {
    # Inform my creator that my server is running.
    set port [lindex $argv 1]
    if {[catch {socket localhost $port} s]} {
        $master(server) Delete
        $master(protocol) Delete
        $master(log) Append "Unable to call back $port: $s"
        error "Unable to call back $port: $s"
    }
    $master(log) Append "Called back $port"
    close $s
}

vwait master(forever)
