# FILE: option.tcl
#
#	An option database for configuring Oc_Classes.
#
# Last modified on: $Date: 1998/06/29 23:16:57 $
# Last modified by: $Author: dgp $
#
# This file defines the Tcl command Oc_Option as the interface to an option
# database for Oc_Classes.  It plays a similar role for Oc_Classes as the
# Tk command [option] plays for Tk windows.

Oc_Class Oc_Option {

    array common db

    ClassConstructor {
        # Source the option configuration file(s)
        set configDir [file join [file dirname [file dirname [file dirname \
                [Oc_DirectPathname [info script]]]]] config]
        set fn [file join $configDir options.tcl]
        if {[file readable $fn]} {
            if {[catch {source $fn} msg]} {
                Oc_Log Log "Error sourcing $fn:
	$msg" warning Oc_Option
            }
        }
    }

    proc Add {app className option value} {

        # Eventually should match the name of the running application
        # against $app and do nothing if the match fails.  There is not 
        # yet any support for determining the name of the running
        # application, so $app is ignored for now.

        if {[regexp , $className]} {
            error "Class pattern may not contain ','"
        }

        lappend db($className,) $option
        set len [llength $db($className,)]
        if {[lsearch -exact $db($className,) $option] + 1 < $len} {
            set db($className,) [lrange $db($className,) 0 [expr {$len-2}]]
        } else {
            set db($className,) [lsort $db($className,)]
        }

        set db($className,$option) $value
        if {![string length $className]} {
            return
        }
        if {[llength [set cmds [info commands $className]]] > 1} {
            error "Ambiguous class pattern '$className': $cmds"
        }
        if {[llength $cmds]} {
            set cmd [lindex $cmds 0]
            if {![catch {$cmd Configure -$option}]} {
                $cmd Configure -$option $value
            }
            return
        }

        # Might want to set an eventHandler here to be triggered by the
        # creation of the Oc_Class $cmd which would call its Configure
        # proc.  Currently no such event is generated.  Also, one should
        # be careful to register only one handler per option.  Even better
        # would be to merge all handlers into one.
        #
        # For now, queries of Oc_Option are hard coded into the 
        # implementation of the Oc_Class command.

    }

    proc Get {className args} {
        set retList {}
        if {[llength $args] == 0} {
            foreach c [array names db $className,] {
                regsub ,$ $c {} c
                lappend retList $c
            }
            return $retList
        }
        set option [lindex $args 0]
        if {[llength $args] == 1} {
            if {![info exists db($className,)]} {
                return $retList
            }
            foreach o $db($className,) {
                if {[string match $option $o]} {
                    lappend retList $o
                }
            }
            return $retList
        }
        set varName [lindex $args 1]
        if {[llength $args] == 2} {
            upvar $varName v
            return [catch {set v $db($className,$option)}]
        }
        return -code error "Usage: $class Get class ?option? ?varName?"
    }

    Constructor {} {
        error "Instances of $class are not allowed"
    }

}

