/* FILE: vecfile.cc                 -*-Mode: c++-*-
 *
 * Main source file for vector field file I/O.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "nb.h"

#include "fileio.h"  // Obsoleted *.vio input
#include "vecfile.h"

/* End includes */     // This is an optional directive to build.tcl,
                       // that excludes the remainder of the file from
                       // '#include ".*"' dependency building

///////////////////////////////////////////////////////////////////////////
// Vf_FileInputID Class

const ClassDoc Vf_FileInputID::class_doc("Vf_FileInputID",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","18-Mar-1997");


Vf_FileInputID::Vf_FileInputID(const char *new_filetype,
                               BOOL (*newpIsType)(const char *,FILE *,
                                                  Tcl_Channel,const char *),
                               Vf_FileInput* (*newpMakeNew)(const char *,
                                                            FILE *,
                                                            Tcl_Channel,
                                                            const char *),
                               BOOL (*pRegister)(const Vf_FileInputID &))
  : filetype(new_filetype)
{
#define MEMBERNAME "Vf_FileInputID"
  pIsType=newpIsType;
  pMakeNew=newpMakeNew;
  if(pRegister!=NULL)  (*pRegister)(*this);
#undef MEMBERNAME
}

void Vf_FileInputID::Dup(const Vf_FileInputID &rhs)
{ // Assignment operator
#define MEMBERNAME "Dup"
  filetype=rhs.filetype;
  pIsType=rhs.pIsType;
  pMakeNew=rhs.pMakeNew;
#undef MEMBERNAME
}

Vf_FileInputID& Vf_FileInputID::operator=(
			   const Vf_FileInputID &rhs)
{ // Assignment operator
  Dup(rhs);
  return *this;
}

Vf_FileInputID::Vf_FileInputID(const Vf_FileInputID
					       &copyobj) // Copy constructor
{
  Dup(copyobj);
}

Vf_FileInputID::Vf_FileInputID()
{ // Default constructor; Needed by Nb_List<T>
#define MEMBERNAME "Vf_FileInputID()"
  pIsType=NULL;
  pMakeNew=NULL;
#undef MEMBERNAME
}

Vf_FileInputID::~Vf_FileInputID()
{ // Dummy function for now
#define MEMBERNAME "~Vf_FileInputID"
#undef MEMBERNAME
}

///////////////////////////////////////////////////////////////////////////
// Vf_FileInput Class

const ClassDoc Vf_FileInput::class_doc("Vf_FileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","17-Mar-1997");

Nb_List<Vf_FileInputID> Vf_FileInput::child_list;

Vf_FileInput::Vf_FileInput(const char *input_filename,
                           FILE *new_infile,
                           Tcl_Channel new_channel,
                           const char *new_channelName)
  : filename(input_filename),infile(new_infile),
    channel(new_channel),channelName(new_channelName)
{
#define MEMBERNAME "Vf_FileInput"
#undef MEMBERNAME
}

Vf_FileInput::~Vf_FileInput()
{
#define MEMBERNAME "~Vf_FileInput"
  if(infile!=NULL)  fclose(infile);
  Tcl_Interp* interp=Oc_GlobalInterpreter();
  if(channel!=NULL && interp!=NULL)
    Tcl_UnregisterChannel(interp, channel);
  // NOTE: "channel" should be registered with the global Tcl
  // interpreter, and should therefore be automatically
  // closed when "unregistered."
#undef MEMBERNAME
}

// All children classes of VFFI should register themselves with
// RegisterChild so that they can be called as necessary from
// NewReader().
BOOL Vf_FileInput::RegisterChild(const Vf_FileInputID &id)
{
#define MEMBERNAME "RegisterChild"
  Vf_FileInputID *id_ptr;
  // Check for repeats
  for(id_ptr=child_list.GetFirst(); id_ptr!=NULL;
      id_ptr=child_list.GetNext()) {
    if(StringCompare(id.filetype,id_ptr->filetype)==0)
      FatalError(-1,STDDOC,"Type %s already registered",
		 (const char*)id_ptr->filetype);
  }
  child_list.Append(id);
  return 0;
#undef MEMBERNAME
}


// Clients should call NewReader() to get a VFFI instance.  The
// client is responsible for destroying (via delete) the returned
// object, i.e., treat this as a "new" operator.
Vf_FileInput *Vf_FileInput::NewReader(const char *new_file)
{
#define MEMBERNAME "NewReader"
  // Open to both a FILE* and a Tcl_Channel
  FILE *fptr=Nb_FOpen(new_file,"rb");
  if(fptr==NULL) {
    ClassMessage(STDDOC,"Unable to open file %s for reading",new_file);
    return NULL;
  }
  Tcl_Interp* interp=Oc_GlobalInterpreter();
  if(interp==NULL)
    FatalError(-1,STDDOC,"Tcl interpretor not initialized");
  char *tmpbuf=new char[strlen(new_file)+16];
  strcpy(tmpbuf,new_file);
  Tcl_Channel mychannel=Tcl_OpenFileChannel(interp,tmpbuf,
					    Oc_AutoBuf("r"),0);
  delete[] tmpbuf;
  if(mychannel==NULL) {
    ClassMessage(STDDOC,"Unable to open file %s for Tcl reading",
		 new_file);
    fclose(fptr);
    return NULL;
  }
  // Note: I had intended to just associate "mychannel" with the input
  // stream "fptr" via a call to Tcl_MakeFileChannel, but I ran into several
  // problems: 1) there appear likely to be portability difficulties between
  // Unix and Windows, 2)  the Tcl_MakeFileChannel interface is different in
  // Tcl 8.0 than in 7.x (though this can be handled with a simple #ifdef),
  // 3) Tcl closes the file itself when the file is "unregistered", but
  // Tcl_MakeFileChannel takes a file descriptor, *not* a FILE* stream pointer,
  // so I don't know if the stream gets properly closed (and "fclose"-ing a
  // file more than once has undefined behavior, and in particular crashes
  // on Linux/Axp, and 4) the file descriptor, fileno(fptr), is passed as
  // type ClientData, which is a void*, so you need a call like
  //
  //  int handle=fileno(fptr);
  //  Tcl_Channel mychannel=Tcl_MakeFileChannel((void *)handle,NULL,
  //                                            TCL_READABLE);
  //
  // (_not_ ...((void *)(&handle)).  This causes the "int handle" to be
  // converted to a void*, which then gets converted back to an int on the
  // other end.  I'm a little skittish about doing this on a machines where
  // void* and int are different sizes.
  //   So, I decided it is probably safer to just open the fool file twice.
  // -mjd, 30-Aug-1997.

  Tcl_RegisterChannel(interp, mychannel);
  char *mychannelName=Tcl_GetChannelName(mychannel);

  // Determine file type
  Vf_FileInputID *id_ptr;
  for(id_ptr=child_list.GetFirst(); id_ptr!=NULL;
      id_ptr=child_list.GetNext()) {
    rewind(fptr);  Tcl_Seek(mychannel,0,SEEK_SET);  // Rewind both streams
    if((id_ptr->pIsType)(new_file,fptr,mychannel,mychannelName)) break;
  }

  if(id_ptr==NULL) {
    ClassMessage(STDDOC,"File %s is of unknown type",new_file);
    if(fptr!=NULL)      fclose(fptr);
    if(mychannel!=NULL) Tcl_UnregisterChannel(interp, mychannel);
    return NULL;
  }

  rewind(fptr);  Tcl_Seek(mychannel,0,SEEK_SET);  // Rewind both streams

  return (id_ptr->pMakeNew)(new_file,fptr,mychannel,mychannelName);
  // Note: It is the responsibility of the VFFI destructor to
  //       fclose fptr and unregister mychannel.
#undef MEMBERNAME
}

//////////////////////////////////////////////////////////////////////
// Ovf File Format specs
const ClassDoc Vf_OvfFileFormatSpecs::class_doc("Vf_OvfFileFormatSpecs",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.1","5-Oct-1997");

const char* Vf_OvfFileFormatSpecs::IDString="#oommf";
/// ID string at top of file. The code in CheckIDString assumes
/// this is all lowercase.

// 4- & 8-byte real data block check values
const REAL4 Vf_OvfFileFormatSpecs::CheckValue4=1234567.;
const REAL8 Vf_OvfFileFormatSpecs::CheckValue8=123456789012345.;


BOOL Vf_OvfFileFormatSpecs::CheckIDString(const char *test)
{
  char *buf=new char[strlen(test)+1];
  
  // Strip spaces and convert to lowercase
  char *cptr1=buf;
  const char *cptr2=test;
  while(*cptr2!='\0') {
    if(!isspace(*cptr2)) *(cptr1++) = tolower(*cptr2);
    cptr2++;
  }
  *cptr1='\0';

  BOOL result=(strncmp(buf,IDString,strlen(IDString))==0);

  delete[] buf;
  return result;
}


// Binary I/O routines that perform byteswapping as needed.
// Return 0 on success, !=0 on error.
INT4m Vf_OvfFileFormatSpecs::ReadBinary(FILE *infile,REAL4 *buf,
				     INT4m count)
{
  if((INT4m)fread((char *)buf,sizeof(REAL4),count,infile)!=count)
    return 1;
#if (BYTEORDER == 4321)  // Flip bytes from MSB to LSB on input
  swap_buf(buf,sizeof(REAL4),count);
#endif
  return 0;
}

INT4m Vf_OvfFileFormatSpecs::ReadBinary(FILE *infile,REAL8 *buf,
				     INT4m count)
{
  if((INT4m)fread((char *)buf,sizeof(REAL8),count,infile)!=count)
    return 1;
#if (BYTEORDER == 4321)  // Flip bytes from MSB to LSB on input
  swap_buf(buf,sizeof(REAL8),count);
#endif
  return 0;
}

INT4m Vf_OvfFileFormatSpecs::WriteBinary(FILE *outfile,REAL4 *buf,
				      INT4m count)
{
  char *cptr=(char *)buf;
#if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB on output
  swap_buf(cptr,sizeof(REAL4),count);
#endif
  if((INT4m)fwrite(cptr,sizeof(REAL4),count,outfile)!=count)
    return 1;
  return 0;
}

INT4m Vf_OvfFileFormatSpecs::WriteBinary(FILE *outfile,REAL8 *buf,
				      INT4m count)
{
  char *cptr=(char *)buf;
#if (BYTEORDER == 4321)  // Flip bytes from LSB to MSB on output
  swap_buf(cptr,sizeof(REAL8),count);
#endif
  if((INT4m)fwrite(cptr,sizeof(REAL8),count,outfile)!=count)
    return 1;
  return 0;
}




//////////////////////////////////////////////////////////////////////
// Vf_OvfSegmentHeader
//
const ClassDoc Vf_OvfSegmentHeader::class_doc("Vf_OvfSegmentHeader",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","24-Mar-1997");

Vf_OvfSegmentHeader::Vf_OvfSegmentHeader()
{
#define MEMBERNAME "Vf_OvfSegmentHeader"
  SetBools(0);
#undef MEMBERNAME
}

Vf_OvfSegmentHeader::~Vf_OvfSegmentHeader()
{
}

void Vf_OvfSegmentHeader::SetBools(BOOL newval)
{ // Sets all BOOL's to newval
#define MEMBERNAME "SetBools"
  title_bool=desc_bool=
    meshunit_bool=xbase_bool=ybase_bool=zbase_bool=
    xstepsize_bool=ystepsize_bool=zstepsize_bool=
    xnodes_bool=ynodes_bool=znodes_bool=
    xmin_bool=ymin_bool=zmin_bool=
    xmax_bool=ymax_bool=zmax_bool=
    valueunit_bool=valuemultiplier_bool=
    valuerangemaxmag_bool=valuerangeminmag_bool=
    meshtype_bool=pointcount_bool=bdryline_bool=newval;
#undef MEMBERNAME
}

BOOL Vf_OvfSegmentHeader::CompleteSet(void)
{ // Returns 1 iff a valid "complete" set of the BOOLS are set.
#define MEMBERNAME "CompleteSet"
  // Currently 2 cases: rectangular & irregular grids.
  // The only fields optional for all files are desc & bdryline.
  BOOL set_code=1;

  // The following fields are required for all ovf files
  set_code=title_bool*meshunit_bool
    *xmin_bool*ymin_bool*zmin_bool*xmax_bool*ymax_bool*zmax_bool
    *valueunit_bool*valuemultiplier_bool
    *valuerangemaxmag_bool*valuerangeminmag_bool
    *meshtype_bool;
  if(!set_code) return set_code;

  if(meshtype==vf_oshmt_rectangular) {
    // Default to rectangular grid, for backwards compatibility
    // with older ovf files that don't include the meshtype data field
    set_code*=xbase_bool*ybase_bool*zbase_bool
      *xstepsize_bool*ystepsize_bool*zstepsize_bool
      *xnodes_bool*ynodes_bool*znodes_bool;
  }
  else if(meshtype==vf_oshmt_irregular) {
    set_code*=pointcount_bool;
  }
  else {
    ClassMessage(STDDOC,"Unknown meshtype: %d\n",(int)meshtype);
    set_code=0;
  }
  return set_code;
#undef MEMBERNAME
}

void Vf_OvfSegmentHeader::SetFieldValue(const Nb_DString &field,
				     const Nb_DString &value)
{ // Note: The import field should be set into lowercase before calling
  //   this routine.  This routine will duplicate and lowercase value
  //   as needed.
#define MEMBERNAME "Vf_OvfSetFieldValue"
  if(field.Length()==0) return;  // Blank line
  if(strcmp((const char *)field,"title")==0) {
    title=value;    title_bool=1;
  }
  else if(strcmp((const char *)field,"desc")==0) {
    if(desc_bool==0) { desc=value; desc_bool=1; }
    else { desc.Append("\n"); desc.Append(value); } // Separate multiple
    /// Desc input records with newlines.
  }
  else if(strcmp((const char *)field,"meshunit")==0) {
    meshunit=value;    meshunit_bool=1;
  }
  else if(strcmp((const char *)field,"meshtype")==0) {
    Nb_DString value_lower(value);
    value_lower.Trim();
    value_lower.ToLower();
    if(strcmp((const char *)value_lower,"rectangular")==0) {
      meshtype=vf_oshmt_rectangular;   meshtype_bool=1;
    }
    else if(strcmp((const char *)value_lower,"irregular")==0) {
      meshtype=vf_oshmt_irregular;     meshtype_bool=1;
    }
    else {
      ClassMessage(STDDOC,"Unknown ovf mesh type value: %s (skipping)",
		   (const char *)value);
    }
  }
  else if(strcmp((const char *)field,"xbase")==0) {
    xbase=Nb_Atof((const char *)value); xbase_bool=1;
  }
  else if(strcmp((const char *)field,"ybase")==0) {
    ybase=Nb_Atof((const char *)value); ybase_bool=1;
  }
  else if(strcmp((const char *)field,"zbase")==0) {
    zbase=Nb_Atof((const char *)value); zbase_bool=1;
  }
  else if(strcmp((const char *)field,"xstepsize")==0) {
    xstepsize=Nb_Atof((const char *)value); xstepsize_bool=1;
  }
  else if(strcmp((const char *)field,"ystepsize")==0) {
    ystepsize=Nb_Atof((const char *)value); ystepsize_bool=1;
  }
  else if(strcmp((const char *)field,"zstepsize")==0) {
    zstepsize=Nb_Atof((const char *)value); zstepsize_bool=1;
  }
  else if(strcmp((const char *)field,"xnodes")==0) {
    xnodes=atoi((const char *)value); xnodes_bool=1;
  }
  else if(strcmp((const char *)field,"ynodes")==0) {
    ynodes=atoi((const char *)value); ynodes_bool=1;
  }
  else if(strcmp((const char *)field,"znodes")==0) {
    znodes=atoi((const char *)value); znodes_bool=1;
  }
  else if(strcmp((const char *)field,"pointcount")==0) {
    pointcount=atoi((const char *)value); pointcount_bool=1;
  }
  else if(strcmp((const char *)field,"xmin")==0) {
    xmin=Nb_Atof((const char *)value); xmin_bool=1;
  }
  else if(strcmp((const char *)field,"ymin")==0) {
    ymin=Nb_Atof((const char *)value); ymin_bool=1;
  }
  else if(strcmp((const char *)field,"zmin")==0) {
    zmin=Nb_Atof((const char *)value); zmin_bool=1;
  }
  else if(strcmp((const char *)field,"xmax")==0) {
    xmax=Nb_Atof((const char *)value); xmax_bool=1;
  }
  else if(strcmp((const char *)field,"ymax")==0) {
    ymax=Nb_Atof((const char *)value); ymax_bool=1;
  }
  else if(strcmp((const char *)field,"zmax")==0) {
    zmax=Nb_Atof((const char *)value); zmax_bool=1;
  }
  else if(strcmp((const char *)field,"boundary")==0) {
    char *work=new char[value.Length()+1];
    strcpy(work,(const char *)value);
    Nb_Vec3<REAL4> point;
    char *cptr=strtok(work," \t");
    while(cptr!=NULL && *cptr!='#') {
      // Load points into boundary list
      point.x=Nb_Atof(cptr);
      cptr=strtok(NULL," \t");
      if(cptr==NULL || *cptr=='#') {
	ClassMessage(STDDOC,"Incomplete boundary point (skipping)");
	break;
      }
      point.y=Nb_Atof(cptr);
      cptr=strtok(NULL," \t");
      if(cptr==NULL || *cptr=='#') {
	ClassMessage(STDDOC,"Incomplete boundary point (skipping)");
	break;
      }
      point.z=Nb_Atof(cptr);
      bdryline.Append(point);
      cptr=strtok(NULL," \t");
    }
    bdryline_bool=1;
  }
  else if(strcmp((const char *)field,"valueunit")==0) {
    valueunit=value;   valueunit_bool=1;
  }
  else if(strcmp((const char *)field,"valuemultiplier")==0) {
    valuemultiplier=Nb_Atof((const char *)value); valuemultiplier_bool=1;
  }
  else if(strcmp((const char *)field,"valuerangemaxmag")==0) {
    valuerangemaxmag=Nb_Atof((const char *)value); valuerangemaxmag_bool=1;
  }
  else if(strcmp((const char *)field,"valuerangeminmag")==0) {
    valuerangeminmag=Nb_Atof((const char *)value); valuerangeminmag_bool=1;
  }
  else {
    ClassMessage(STDDOC,"Unknown ovf segment header field: %s (skipping)",
		  (const char *)field);
  }
#undef MEMBERNAME
}


///////////////////////////////////////////////////////////////////////////
// Vf_OvfFileInput Class

const ClassDoc Vf_OvfFileInput::class_doc("Vf_OvfFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.1","25-Jul-1997");

const char* Vf_OvfFileInput::file_type="OOMMF (.ovf) Input File";

const Vf_FileInputID Vf_OvfFileInput::id(Vf_OvfFileInput::file_type,
                                         Vf_OvfFileInput::IsType,
                                         Vf_OvfFileInput::MakeNew,
                                         Vf_FileInput::RegisterChild);

Vf_OvfFileInput::Vf_OvfFileInput(const char *new_filename,
                                 FILE *new_infile,
                                 Tcl_Channel new_channel,
                                 const char *new_channelName)
  : Vf_FileInput(new_filename,new_infile,new_channel,new_channelName)
{
#define MEMBERNAME "Vf_OvfFileInput"
  // Read file header here (i.e., determine # of segments)
#undef MEMBERNAME
}

Vf_FileInput* Vf_OvfFileInput::MakeNew(const char *new_filename,
					    FILE *new_infile,
					    Tcl_Channel new_channel,
					    const char *new_channelName)
{
  return new Vf_OvfFileInput(new_filename,new_infile,
			  new_channel,new_channelName);
}

BOOL Vf_OvfFileInput::IsType(const char *,
			  FILE *test_fptr,
			  Tcl_Channel,
			  const char *)
{
#define MEMBERNAME "IsType"
  char buf[SCRATCH_BUF_SIZE];
  buf[0]='\0'; // Safety
  fgets(buf,sizeof(buf),test_fptr);

  // Do string comparison
  return Vf_OvfFileFormatSpecs::CheckIDString(buf);

#undef MEMBERNAME
}

INT4m Vf_OvfFileInput::ReadLine(Nb_DString &buf)
{ // Reads one line from infile and sticks it into buf
  // Returns 0 on success, 1 on EOF or other error
  // NOTE: This routine is not terribly efficient (lots of
  //  unnecessary calls to string allocation/deallocation),
  //  and should probably only be used on ovf file headers
  // (*not* data sections).
#define MEMBERNAME "ReadLine"
  static char readbuf[SCRATCH_BUF_SIZE];
  char *cptr;
  buf="";  // Clear buf;
  do {
    readbuf[SCRATCH_BUF_SIZE-2]='\0';  // Used for line-too-long check
    cptr=fgets(readbuf,SCRATCH_BUF_SIZE,infile);
    if(cptr==NULL) return 1; // Error
    buf.Append(readbuf,SCRATCH_BUF_SIZE);
  } while(readbuf[SCRATCH_BUF_SIZE-2]!='\0'
	&& readbuf[SCRATCH_BUF_SIZE-2]!='\n');
  return 0;
#undef MEMBERNAME
}

INT4m Vf_OvfFileInput::ParseContentLine(const Nb_DString &buf,
				     Nb_DString &field,Nb_DString &value)
{ //   Breaks buf (import) down into field & value substrings (exports),
  // with field put into canonical form (strips whitespace & underscores,
  // and converts to lowercase).
  //   Return 0 on success, 1 on error.  The "field" value will be
  // a null string ('\0') if buf is an empty or comment line.
#define MEMBERNAME "ParseContentLine"
  field="";  value="";  // Default return values

  // Copy buf into workspace, trimming leading & trailing whitespace
  Nb_DString workstr(buf);
  workstr.Trim();
  char *work=&(workstr[0]); // This is cheating!  We can't use workstr
  // again from below this point, since if we do *work may become invalid.
  // It has the advantage, though, that the memory allocated by workstr
  // is automatically freed upon function exit.

  // Tokenize workspace
  char *fieldstart=strchr(work,'#');
  if(fieldstart==NULL) return 0;  // Possibly illegal line, but assume blank
  fieldstart++;
  if(*fieldstart=='#') return 0;  // Comment line; treat as blank
  char *valuestart=strchr(fieldstart,':');
  if(valuestart==NULL) return 0;  // Possibly illegal line, but assume blank
  *valuestart='\0'; valuestart++;

  // Collapse field string (remove all non alpha-numeric characters) and
  // convert to lowercase.
  char *cptr1,*cptr2;
  cptr1=cptr2=fieldstart;
  while(*cptr2!='\0') {
    if(isalnum(*cptr2)) *(cptr1++) = tolower(*cptr2);
    cptr2++;
  }
  *cptr1='\0';
  field=fieldstart;

  // Skip leading whitespace in value string
  while(*valuestart!='\0' && isspace(*valuestart)) valuestart++;

  // If field is 'desc', then return as is.
  if(strncmp("desc",fieldstart,strlen("desc"))==0) {
    value=valuestart;
    return 0;
  }

  // Otherwise strip trailing comment (if any).
  cptr1=valuestart;
  while((cptr1=strchr(cptr1,'#'))!=NULL) if(*(++cptr1)=='#') break;
  if(cptr1!=NULL) *(cptr1-1)='\0';

  // Strip trailing whitespace
  cptr1=strchr(valuestart,'\0');
  while(cptr1>valuestart && isspace(*(--cptr1))) (*cptr1)='\0';

  // If field is "begin" or "end", then convert value to lowercase
  // and strip whitespace
  if(strcmp(field,"begin")==0 || strcmp(field,"end")==0) {
    cptr1=cptr2=valuestart;
    while(*cptr2!='\0') {
      if(isalnum(*cptr2)) *(cptr1++) = tolower(*cptr2);
      cptr2++;
    }
    *cptr1='\0';
  }

  value=valuestart;
  return 0;
#undef MEMBERNAME
}

// Read 3 ASCII floating point values from infile, skipping
// any leading whitespace and eating the first whitespace
// character after the third value.
// Returns 0 on success, 1 on error.
INT4m Vf_OvfFileInput::ReadTextVec3f(Nb_Vec3<REAL4> &data)
{
#define MEMBERNAME "ReadTextData"
  static char buf[SCRATCH_BUF_SIZE];
  int ch,bufindex;
  int elt=0;  // Next element to fill; 0->x, 1->y, 2->z, 3->done
  REAL4 value;
  while(elt<3) {
    // Skip leading white space and comments
    while(1) {
      ch=fgetc(infile);
      if(ch==EOF) return 1;  // Premature end-of-file
      // Check for comment.
      if(ch=='#') {
	ch=fgetc(infile);
	if(ch==EOF) return 1;  // Premature end-of-file
	if(ch=='#') {
	  // Comment; read to end of line
	  while((ch=fgetc(infile))!='\n') {
	    if(ch==EOF) return 1;  // Premature end-of-file
	  }
	} else {
	  // Blank line; read to end of line
	  while ((ch=fgetc(infile))!='\n') {
	    if(ch==EOF) return 1;  // Premature end-of-file
	    if(!isspace(ch)) {
	      ClassMessage(STDDOC,"Field-value pair inside data block,"
			   " or premature end of data in file %s.",
			   (const char *)filename);
	      return 1;
	    }
	  }
	}
	continue;
      }
      // Non-comment white-space check
      if(isspace(ch)) continue;
      // Otherwise, should be data.
      break;
    }

    // Fill buffer with data up to next white space character
    bufindex=0;
    do {
      if(ch=='d' || ch=='D') ch='e';  // Convert FORTRAN 'd' exponent
      buf[bufindex++]=ch;             // designator to C 'e' designator
      ch=fgetc(infile);
      if(ch==EOF) return 1;  // Premature end-of-file
    } while(!isspace(ch) && bufindex<SCRATCH_BUF_SIZE-1);
    if(bufindex>=SCRATCH_BUF_SIZE-1) {
      // Single data elt too big to fit in buffer?!
      ClassMessage(STDDOC,"Bad data in file %s;"
		    " Single value >%d characters long",
		    (const char *)filename,SCRATCH_BUF_SIZE);
      return 1;
    }
    buf[bufindex]='\0';

    // Convert to floating point and store
    char* endptr;
    value=strtod(buf,&endptr);
    if( (endptr-buf) != bufindex ) {
      ClassMessage(STDDOC,"Non-numeric data ->%s<- in data block of file %s",
		    buf,(const char *)filename);
      return 1;
    }
    if(elt==0)       data.x=value;
    else if(elt==1)  data.y=value;
    else             data.z=value;
    elt++;
  }

  return 0;
#undef MEMBERNAME
}

// Read 3 floating point values from infile, using 'datastyle'
//   format.  See notes in Vf_OvfFileInput::ReadTextVec3f for
//   more details on ASCII (text) reads.  Results stored in 'data'.
// Returns 0 on success, 1 on error.
INT4m Vf_OvfFileInput::ReadVec3f(Vf_OvfDataStyle datastyle,
			      Nb_Vec3<REAL4> &data)
{
#define MEMBERNAME "ReadVec3f"
  if(datastyle == vf_oascii) return ReadTextVec3f(data);
  if(datastyle == vf_obin4)      {  // 4-byte binary type
    REAL4 databuf[3];
    if(Vf_OvfFileFormatSpecs::ReadBinary(infile,databuf,3)!=0) return 1;
    data.x=databuf[0]; data.y=databuf[1]; data.z=databuf[2];
  }
  else if(datastyle == vf_obin8) {  // 8-byte binary type
    REAL8 databuf[3];
    if(Vf_OvfFileFormatSpecs::ReadBinary(infile,databuf,3)!=0) return 1;
    data.x=databuf[0]; data.y=databuf[1]; data.z=databuf[2];
  }
  return 0;
#undef MEMBERNAME
}

Vf_Mesh *Vf_OvfFileInput::NewMesh()
{
#define MEMBERNAME "NewMesh"
  Nb_DString buf,field,value;
  Vf_OvfDataStyle datastyle;
  const char *datastylestring=NULL;

  // Skip preamble
  while(1) {
    if(ReadLine(buf)!=0) {
      ClassMessage(STDDOC,"Premature EOF in file %s "
		    "(no segment header found)",
		    (const char *)filename);
      return new Vf_EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"begin")==0  && strcmp(value,"header")==0) break;
  }

  // Read head and store results into Vf_OvfSegmentHeader struct
  Vf_OvfSegmentHeader ovfseghead;
  ovfseghead.SetBools(0);  // Safety
  ovfseghead.SetFieldValue("desc","");  // Description field is optional
  ovfseghead.SetFieldValue("meshtype","rectangular");
  /// Set default meshtype to rectangular, for older ovf files w/o
  /// the meshtype field.
  while(1) {
    if(ReadLine(buf)!=0) {
      ClassMessage(STDDOC,"Premature EOF in file %s "
		    "(missing segment header end mark)",
		    (const char *)filename);
      return new Vf_EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"end")==0 && strcmp(value,"header")==0) break;
    if(field.Length()>0) ovfseghead.SetFieldValue(field,value);
  }
  if(!ovfseghead.CompleteSet()) {
      ClassMessage(STDDOC,"Incomplete segment header in file %s; Skipping",
		    (const char *)filename);
      return new Vf_EmptyMesh();
  }

  // Skip ahead in file to data section
  while(1) {
    if(ReadLine(buf)!=0) {
      ClassMessage(STDDOC,"Premature EOF in file %s (no data section found)",
		    (const char *)filename);
      return new Vf_EmptyMesh();
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"begin")==0) {
      if(strcmp(value,"datatext")==0) {
	// Ascii data
	datastyle=vf_oascii;
	datastylestring="datatext";
	break;
      }
      if(strcmp(value,"databinary4")==0) {
	// 4-byte binary
	datastyle=vf_obin4;
	datastylestring="databinary4";
	break;
      }
      if(strcmp(value,"databinary8")==0) {
	// 8-byte binary
	datastyle=vf_obin8;
	datastylestring="databinary8";
	break;
      }
    }
  }

  // Set up boundary display list
  Nb_BoundingBox<REAL4> total_range;

  if(!ovfseghead.bdryline_bool) {
    // Boundary line not set directly.  Build one from ?min & ?max
    REAL8 zave=(ovfseghead.zmin+ovfseghead.zmax)/2.0;
    ovfseghead.bdryline.SetFirst(Nb_Vec3<REAL4>(ovfseghead.xmin,
						ovfseghead.ymin,zave));
    ovfseghead.bdryline.Append(Nb_Vec3<REAL4>(ovfseghead.xmin,
					      ovfseghead.ymax,zave));
    ovfseghead.bdryline.Append(Nb_Vec3<REAL4>(ovfseghead.xmax,
					      ovfseghead.ymax,zave));
    ovfseghead.bdryline.Append(Nb_Vec3<REAL4>(ovfseghead.xmax,
					      ovfseghead.ymin,zave));
    ovfseghead.bdryline.Append(Nb_Vec3<REAL4>(ovfseghead.xmin,
					      ovfseghead.ymin,zave));
  }

  // Setup total_range bounding box to include the boundary list
  // and the ?min-$max corners.  (Regular grids additionally
  // expand to include all data points.  See "Rectangular grid"
  // initialization section below.)
  Nb_Vec3<REAL4> corner1(ovfseghead.xmin,ovfseghead.ymin,ovfseghead.zmin);
  Nb_Vec3<REAL4> corner2(ovfseghead.xmax,ovfseghead.ymax,ovfseghead.zmax);
  total_range.SortAndSet(corner1,corner2);
  for(Nb_Vec3<REAL4> *vptr=ovfseghead.bdryline.GetFirst();vptr!=NULL;
      vptr=ovfseghead.bdryline.GetNext()) {
    total_range.ExpandWith(*vptr);
  }

  // Determine data normalization value
  double norm=ovfseghead.valuerangemaxmag;
  if(norm>SQRT_REAL8_EPSILON) norm=1./norm;
  else norm=1.0;

  // Initialize mesh
  Vf_GridVec3f *mesh_rect=NULL;
  Vf_GeneralMesh3f *mesh_irreg=NULL;
  Vf_Mesh *mesh;
  if(ovfseghead.meshtype!=vf_oshmt_irregular) {
    // Rectangular grid
    Nb_Vec3<REAL4> basept(ovfseghead.xbase,ovfseghead.ybase,ovfseghead.zbase);
    Nb_Vec3<REAL4> lastpt=basept+Nb_Vec3<REAL4>(
                              (ovfseghead.xnodes-1)*ovfseghead.xstepsize,
			      (ovfseghead.ynodes-1)*ovfseghead.ystepsize,
			      (ovfseghead.znodes-1)*ovfseghead.zstepsize);
    total_range.ExpandWith(basept); // Expand to include all datapoints.
    total_range.ExpandWith(lastpt);
    Nb_BoundingBox<REAL4> data_range; data_range.SortAndSet(basept,lastpt);

    Nb_Vec3<REAL4> gridstep(ovfseghead.xstepsize,ovfseghead.ystepsize,
		   ovfseghead.zstepsize);
    mesh_rect=new Vf_GridVec3f((const char *)filename,
			       (const char *)ovfseghead.title,
			       (const char *)ovfseghead.desc,
			       (const char *)ovfseghead.valueunit,
			       ovfseghead.valuemultiplier/norm,
			       ovfseghead.xnodes,
			       ovfseghead.ynodes,ovfseghead.znodes,
			       basept,gridstep,ovfseghead.bdryline,
			       data_range,total_range);
    mesh=mesh_rect;
  }
  else {
    // Irregular grid
    mesh_irreg=new Vf_GeneralMesh3f((const char *)filename,
                                    (const char *)ovfseghead.title,
                                    (const char *)ovfseghead.desc,
				    (const char *)ovfseghead.valueunit,
				    ovfseghead.valuemultiplier/norm);
    mesh_irreg->SetRange(total_range);
    mesh_irreg->SetBoundaryList(ovfseghead.bdryline);
    if(ovfseghead.xstepsize_bool) {
      // Set step hints
      Nb_Vec3<REAL4> step;
      step.x=ovfseghead.xstepsize;
      if(!ovfseghead.ystepsize_bool) step.y=ovfseghead.xstepsize;
      else                           step.y=ovfseghead.ystepsize;
      if(!ovfseghead.zstepsize_bool) step.z=ovfseghead.xstepsize;
      else                           step.z=ovfseghead.zstepsize;
      mesh_irreg->SetStepHints(step);
    }
    mesh=mesh_irreg;
  }

  // If binary data, verify check value
  if(datastyle == vf_obin4) {
    REAL4 test=0.;
    Vf_OvfFileFormatSpecs::ReadBinary(infile,&test,1);
    if(!Vf_OvfFileFormatSpecs::CheckValue(test)) {
      ClassMessage(STDDOC,
	    "Invalid data block check value in file %s; Wrong byte order?",
		    (const char *)filename);
      delete mesh;
      return new Vf_EmptyMesh();
    }
  }
  else if(datastyle == vf_obin8) {
    REAL8 test;
    Vf_OvfFileFormatSpecs::ReadBinary(infile,&test,1);
    if(!Vf_OvfFileFormatSpecs::CheckValue(test)) {
      ClassMessage(STDDOC,
	    "Invalid data block check value in file %s; Wrong byte order?",
		    (const char *)filename);
      delete mesh;
      return new Vf_EmptyMesh();
    }
  }

  // Read data
  if(ovfseghead.meshtype!=vf_oshmt_irregular) {
    // Rectangular grid
    INT4m i,j,k;
    Nb_Vec3<REAL4> data;
    for(k=0;k<ovfseghead.znodes;k++)
      for(j=0;j<ovfseghead.ynodes;j++)
	for(i=0;i<ovfseghead.xnodes;i++) {
	  // Read one entry
	  if(ReadVec3f(datastyle,data)!=0) {
	    ClassMessage(STDDOC,"Premature EOF in file %s in data section",
			  (const char *)filename);
	    delete mesh;
	    return new Vf_EmptyMesh();
	  }
	  data*=norm; // Normalize data.
	  (*mesh_rect)(i,j,k)=data;
	}
  }
  else {
    // Irregular grid
    INT4m i;
    Nb_LocatedVector<REAL4> lv;
    for(i=0;i<ovfseghead.pointcount;i++) {
      // Read one entry
      if(ReadVec3f(datastyle,lv.location)!=0
	 || ReadVec3f(datastyle,lv.value)!=0) {
	ClassMessage(STDDOC,"Premature EOF in file %s in data section",
		      (const char *)filename);
	delete mesh;
	return new Vf_EmptyMesh();
      }
      lv.value*=norm; // Normalize data.
      mesh_irreg->AddPoint(lv);
    }     
    mesh_irreg->SortPoints();
  }

  // Read data block end statement
  while(1) {
    if(ReadLine(buf)!=0) {
      ClassMessage(STDDOC,"Premature EOF in file %s (no data end marker)",
		    (const char *)filename);
      return mesh;
    }
    ParseContentLine(buf,field,value);
    if(field[0]=='\0') continue; // Skip blank lines
    if(strcmp(field,"end")!=0) {
      ClassMessage(STDDOC,"Missing data end marker in file %s",
		    (const char *)filename);
      return mesh;
    }
    if(strcmp(value,datastylestring)==0) break; // Success!
    ClassMessage(STDDOC,"Data end marker wrong type in file %s "
		  "(->%s<- instead of ->%s<-)",
		  (const char *)filename,
		  (const char *)value,datastylestring);
    return mesh;
  }

  // Skip to segment end
  while(1) {
    if(ReadLine(buf)!=0) {
      ClassMessage(STDDOC,"Premature EOF in file %s (no segment end marker)",
		    (const char *)filename);
      return mesh;
    }
    ParseContentLine(buf,field,value);
    if(strcmp(field,"end")==0  && strcmp(value,"segment")==0) break;
  }

  // NOTE: *infile is closed by base destructor
  // (Moreover, there may be additional segments in this file.)

  return mesh;
#undef MEMBERNAME
}


///////////////////////////////////////////////////////////////////////////
// Vf_OvfFileOutput Class

const ClassDoc Vf_OvfFileOutput::class_doc("Vf_OvfFileOutput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.1","5-Oct-1997");

INT4m Vf_OvfFileOutput::WriteMesh(Vf_Mesh *mesh,FILE* fptr,
                                  Vf_OvfDataStyle datastyle,
                                  BOOL force_irreg,
                                  const char *title,
                                  const char *desc)
{ // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.
  // If force_irreg is set, the output will be in the Ovf irregular
  // grid file format (location & value) even if input data lies
  // on a rectangular grid.

  enum { mt_rect, mt_irreg } meshtype=mt_rect;
  INT4m xdim,ydim,zdim,pointcount=0;
  Nb_BoundingBox<REAL4> range;

  if(strcmp(mesh->GetMeshType(),"Vf_GridVec3f")==0) {
    meshtype=mt_rect;
  }
  else if(strcmp(mesh->GetMeshType(),"Vf_GeneralMesh3f")==0) {
    meshtype=mt_irreg;
  }
  else {
    // Unsupported mesh type
    return -1;
  }

  // Downcast mesh
  Vf_GridVec3f *mesh_rect=NULL;
  Vf_GeneralMesh3f *mesh_irreg=NULL;
  switch(meshtype)
    {
    case mt_rect:
      mesh_rect=(Vf_GridVec3f *)mesh;
      pointcount=mesh_rect->GetSize();
      if(force_irreg) fprintf(fptr,"# OOMMF: irregular mesh v0.99\n");
      else          fprintf(fptr,"# OOMMF: rectangular mesh v0.99\n");
      break;
    case mt_irreg:
    default:  // Safety
      mesh_irreg=(Vf_GeneralMesh3f *)mesh;
      pointcount=mesh_irreg->GetSize();
      fprintf(fptr,"# OOMMF: irregular mesh v0.99\n");
      break;
    }

  // Write header
  fprintf(fptr,"# Segment count: 1\n");
  fprintf(fptr,"# Begin: Segment\n");
  fprintf(fptr,"# Begin: Header\n");
  if(title==NULL) fprintf(fptr,"# Title: %s\n",mesh->GetTitle());
  else            fprintf(fptr,"# Title: %s\n",title);
  if(desc==NULL) desc=mesh->GetDescription();
  if(desc==NULL || desc[0]=='\0') {
    fprintf(fptr,"# Desc: Ovf file created by class %s v%s (%s)\n",
	    class_doc.classname,class_doc.revision,class_doc.revdate);
  }
  else {
    // Print out description, breaking at newlines as necessary.
    const char *cptr1,*cptr2;
    cptr1=desc;
    while((cptr2=strchr(cptr1,'\n'))!=NULL) {
      fprintf(fptr,"# Desc: %.*s\n",(int)(cptr2-cptr1),cptr1);
      cptr1=cptr2+1;
    }
    if(*cptr1!='\0') fprintf(fptr,"# Desc: %s\n",cptr1);
  }

  if(meshtype==mt_rect && !force_irreg) {
    Nb_Vec3<REAL4> basept=mesh_rect->GetBasePoint();
    Nb_Vec3<REAL4> gridstep=mesh_rect->GetGridStep();  
    mesh_rect->GetDimens(xdim,ydim,zdim);
    fprintf(fptr,"# meshtype: rectangular\n");
    fprintf(fptr,"# meshunit: unknown\n");
    fprintf(fptr,"# xbase: %g\n# ybase: %g\n# zbase: %g\n",
	    basept.x,basept.y,basept.z);
    fprintf(fptr,"# xstepsize: %g\n# ystepsize: %g\n# zstepsize: %g\n",
	    gridstep.x,gridstep.y,gridstep.z);
    fprintf(fptr,"# xnodes: %d\n# ynodes: %d\n# znodes: %d\n",
	    xdim,ydim,zdim);
  }
  else {
    Nb_Vec3<REAL4> stephints;
    if(meshtype==mt_irreg)   stephints=mesh_irreg->GetStepHints();
    else if(meshtype==mt_rect) stephints=mesh_rect->GetGridStep();
    fprintf(fptr,"# meshtype: irregular\n");
    fprintf(fptr,"# meshunit: unknown\n");
    fprintf(fptr,"# pointcount: %d\n",pointcount);
    fprintf(fptr,"# xstepsize: %g\n# ystepsize: %g\n# zstepsize: %g\n",
	    stephints.x,stephints.y,stephints.z);
  }

  Nb_Vec3<REAL4> minpt,maxpt;
  mesh->GetRange(range);
  range.GetExtremes(minpt,maxpt);
  fprintf(fptr,"# xmin: %g\n# ymin: %g\n# zmin: %g\n",
	  minpt.x,minpt.y,minpt.z);
  fprintf(fptr,"# xmax: %g\n# ymax: %g\n# zmax: %g\n",
	  maxpt.x,maxpt.y,maxpt.z);

  Nb_List< Nb_Vec3<REAL4> > bdryline;
  mesh->GetBoundaryList(bdryline);
  if(bdryline.GetSize()>0) {
    fprintf(fptr,"# boundary:");
    for(Nb_Vec3<REAL4> *v=bdryline.GetFirst();v!=NULL;v=bdryline.GetNext()) {
      fprintf(fptr," %g %g %g",v->x,v->y,v->z);
    }
    fprintf(fptr,"\n");
  }

  // In the future, we may want to change the following:
  fprintf(fptr,"# valueunit: %s\n",mesh->GetValueUnit());
  fprintf(fptr,"# valuemultiplier: %-17g\n",mesh->GetValueMultiplier());
  fprintf(fptr,"# ValueRangeMaxMag:  1.\n");
  fprintf(fptr,"# ValueRangeMinMag:  1e-8\n");

  fprintf(fptr,"# End: Header\n");

  // Write data header
  if(datastyle==vf_obin4)     {
    REAL4 check;
    Vf_OvfFileFormatSpecs::GetCheckValue(check);
    fprintf(fptr,"# Begin: data binary 4\n");
    Vf_OvfFileFormatSpecs::WriteBinary(fptr,&check,1);
  }
  else if(datastyle==vf_obin8) {
    REAL8 check;
    Vf_OvfFileFormatSpecs::GetCheckValue(check);
    fprintf(fptr,"# Begin: data binary 8\n");
    Vf_OvfFileFormatSpecs::WriteBinary(fptr,&check,1);
  }
  else if(datastyle==vf_oascii) { 
    fprintf(fptr,"# Begin: data text\n");
  }

  // Write data
  if(meshtype==mt_rect && !force_irreg) {
    INT4m i,j,k;
    Nb_Vec3<REAL4> *v;
    for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
      v=&(mesh_rect->GridVec(i,j,k));
      if(datastyle==vf_obin4) {
	REAL4 buf[3];
	buf[0]=v->x; buf[1]=v->y; buf[2]=v->z;
	Vf_OvfFileFormatSpecs::WriteBinary(fptr,buf,3);
      }
      else if(datastyle==vf_obin8) {
	REAL8 buf[6]; 
	buf[0]=v->x; buf[1]=v->y; buf[2]=v->z;
	Vf_OvfFileFormatSpecs::WriteBinary(fptr,buf,3);
      }
      else if(datastyle==vf_oascii) {
	fprintf(fptr,"%.6g %.6g %.6g\n",v->x,v->y,v->z);
      }
    }
  }
  else {
    // Get all points in mesh
    Nb_List<Vf_DisplayVector> points;
    REAL4m xstep=0.,ystep=0.,zstep=0.;
    mesh->GetDisplayList(xstep,ystep,zstep,range,NULL,points);

    // Dump to file
    Vf_DisplayVector *lv;
    for(lv=points.GetFirst();lv!=NULL;lv=points.GetNext()) {
      if(datastyle==vf_obin4) {
	REAL4 buf[6];
	buf[0]=lv->location.x; buf[1]=lv->location.y; buf[2]=lv->location.z;
	buf[3]=lv->value.x;    buf[4]=lv->value.y;    buf[5]=lv->value.z;
	Vf_OvfFileFormatSpecs::WriteBinary(fptr,buf,6);
      }
      else if(datastyle==vf_obin8) {
	REAL8 buf[6];
	buf[0]=lv->location.x; buf[1]=lv->location.y; buf[2]=lv->location.z;
	buf[3]=lv->value.x;    buf[4]=lv->value.y;    buf[5]=lv->value.z;
	Vf_OvfFileFormatSpecs::WriteBinary(fptr,buf,6);
      }
      else if(datastyle==vf_oascii) {
	fprintf(fptr,"%.6g %.6g %.6g %.6g %.6g %.6g\n",
		lv->location.x,lv->location.y,lv->location.z,
		lv->value.x,lv->value.y,lv->value.z);
      }
    }
  }

  // Write trailer
  if(datastyle==vf_obin4)       fprintf(fptr,"# End: data binary 4\n");
  else if(datastyle==vf_obin8)  fprintf(fptr,"# End: data binary 8\n");
  else if(datastyle==vf_oascii) fprintf(fptr,"# End: data text\n");
  fprintf(fptr,"# End: segment\n");
  return 0;
}

///////////////////////////////////////////////////////////////////////////
// Vf_VioFileOutput Class

const ClassDoc Vf_VioFileOutput::class_doc("Vf_VioFileOutput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","25-Feb-1998");

INT4m Vf_VioFileOutput::WriteMesh(Vf_Mesh *mesh,FILE* fptr,
                                  const char *desc)
{ // Returns -1 if mesh type not supported,
  //          0 on success,
  //         >0 on other error.

  INT4m i,j,xdim,ydim,zdim;

  if(strcmp(mesh->GetMeshType(),"Vf_GridVec3f")!=0) {
    // Unsupported mesh type
    return -1;
  }

  // Downcast mesh
  Vf_GridVec3f& mesh_rect=(Vf_GridVec3f&)(*mesh);
  mesh_rect.GetDimens(xdim,ydim,zdim);
  if(zdim!=1) {
    // Unsupported mesh type
    return -1;
  }

  // Write header
  struct {
    BYTE    Magic[8];
    REAL8   EndianTest;
    INT4    Nx,Ny,Nz,dimen,dataoffset;
    BYTE    Reserved[36];
    BYTE    Note[448];
  } fh;

  strncpy((char *)fh.Magic,"VecFil\x02",sizeof(fh.Magic));
  fh.EndianTest = 1.234;

  fh.Nx=xdim; fh.Nz=ydim; fh.Ny=zdim;  // Note swap of y & z
  fh.dimen=3;
  fh.dataoffset=sizeof(fh);

  for(i=0;i<(int)sizeof(fh.Reserved);i++) fh.Reserved[i]='\0';
  for(i=0;i<(int)sizeof(fh.Note);i++)     fh.Note[i]='\0';
  if(desc!=NULL) strncpy((char *)fh.Note,desc,sizeof(fh.Note)-1);

  if(fwrite((char *)(&fh),sizeof(fh),1,fptr)!=1) return 1;

  // Write data
  REAL8 arr[3];
  for(i=0;i<xdim;i++) for(j=0;j<ydim;j++) {
    Nb_Vec3<REAL4>& v=mesh_rect(i,j,0);
    arr[0]=v.x; arr[1]=-v.z; arr[2]=v.y;  // Change coord system
    if(fwrite((char *)arr,sizeof(arr[0]),3,fptr)!=3)
      return 1;
  }

  return 0;
}

///////////////////////////////////////////////////////////////////////////
// Vf_VioFileInput Class

const ClassDoc Vf_VioFileInput::class_doc("Vf_VioFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","19-Mar-1997");

const char* Vf_VioFileInput::file_type="VIO (.vio) Input File";

const Vf_FileInputID Vf_VioFileInput::id(Vf_VioFileInput::file_type,
                                         Vf_VioFileInput::IsType,
                                         Vf_VioFileInput::MakeNew,
                                         Vf_FileInput::RegisterChild);

Vf_VioFileInput::Vf_VioFileInput(const char *new_filename,
                                 FILE *new_infile,
                                 Tcl_Channel new_channel,
                                 const char *new_channelName)
  : Vf_FileInput(new_filename,new_infile,new_channel,new_channelName)
{
#define MEMBERNAME "Vf_VioFileInput"
  // Read header here
#undef MEMBERNAME
}

Vf_FileInput* Vf_VioFileInput::MakeNew(const char *new_filename,
                                       FILE *new_infile,
                                       Tcl_Channel new_channel,
                                       const char *new_channelName)
{
  return new Vf_VioFileInput(new_filename,new_infile,
                             new_channel,new_channelName);
}

BOOL Vf_VioFileInput::IsType(const char *,
                             FILE *test_fptr,
                             Tcl_Channel,
                             const char *)
{
#define MEMBERNAME "IsType"
  const char   VecFileMagic02[8] = "VecFil\x02";
  char buf[8]="\0\0\0\0\0\0\0"; // Safety init
  fread(buf,1,8,test_fptr);
  if(memcmp(VecFileMagic02,buf,8)==0) return 1;  // Match!
  return 0;
#undef MEMBERNAME
}

Vf_Mesh *Vf_VioFileInput::NewMesh()
{
#define MEMBERNAME "NewMesh"
  Vf_Mesh *newmesh=NewRectGrid();
  if(newmesh==NULL) newmesh=new Vf_EmptyMesh();
  return newmesh;
#undef MEMBERNAME
}

Vf_GridVec3f *Vf_VioFileInput::NewRectGrid()
{
#define MEMBERNAME "NewRectGrid"
  Vf_VioFile viofile;

  // Open file and get grid dimensions
  if(viofile.SetFilePtr(infile)!=0) return NULL;
  INT4m xsize,ysize,zsize;
  viofile.GetDimensions(xsize,ysize,zsize);
  // Set bounding box and boundary list
  // Note: Vio files sample at cell centers
  Nb_BoundingBox<REAL4> data_range,total_range;
  Nb_List< Nb_Vec3<REAL4> > boundary;
  data_range.Set(Nb_Vec3<REAL4>(0.,0.,0.),
          Nb_Vec3<REAL4>(REAL4(xsize-1.),REAL4(ysize-1.),REAL4(zsize-1.)));
  total_range.Set(Nb_Vec3<REAL4>(-0.5,-0.5,-0.5),
	  Nb_Vec3<REAL4>(REAL4(xsize-0.5),REAL4(ysize-0.5),REAL4(zsize-0.5)));
  boundary.SetFirst(Nb_Vec3<REAL4>(-0.5,-0.5,0.0));
  boundary.Append(Nb_Vec3<REAL4>(REAL4(xsize-0.5),-0.5,0.0));
  boundary.Append(Nb_Vec3<REAL4>(REAL4(xsize-0.5),REAL4(ysize-0.5),0.0));
  boundary.Append(Nb_Vec3<REAL4>(-0.5,REAL4(ysize-0.5),0.0));
  boundary.Append(Nb_Vec3<REAL4>(-0.5,-0.5,0.0));

  // Initialize mesh
  Vf_GridVec3f *mesh=new Vf_GridVec3f(filename,filename,"vio file",
				      NULL,1.0,
                                      xsize,ysize,zsize,
                                      Nb_Vec3<REAL4>(0.,0.,0.),
                                      Nb_Vec3<REAL4>(1.,1.,1.),
                                      boundary,data_range,total_range);

  // Fill grid
  viofile.FillArray(*mesh);

  // NOTE: *infile is closed by base destructor

  return mesh;
#undef MEMBERNAME
}

///////////////////////////////////////////////////////////////////////////
// Vf_SvfFileInput Class

const ClassDoc Vf_SvfFileInput::class_doc("Vf_SvfFileInput",
	      "Michael J. Donahue (michael.donahue@nist.gov)",
	      "1.0.0","30-Aug-1997");

const char* Vf_SvfFileInput::file_type="Svf (.svf) Input File";

const Vf_FileInputID Vf_SvfFileInput::id(Vf_SvfFileInput::file_type,
                                         Vf_SvfFileInput::IsType,
                                         Vf_SvfFileInput::MakeNew,
                                         Vf_FileInput::RegisterChild);

Vf_SvfFileInput::Vf_SvfFileInput(const char *new_filename,
                                 FILE *new_infile,
                                 Tcl_Channel new_channel,
                                 const char *new_channelName)
  : Vf_FileInput(new_filename,new_infile,
                 new_channel,new_channelName)
{
#define MEMBERNAME "Vf_SvfFileInput"
#undef MEMBERNAME
}

Vf_SvfFileInput::~Vf_SvfFileInput()
{
#define MEMBERNAME "~Vf_SvfFileInput"
#undef MEMBERNAME
}

Vf_FileInput* Vf_SvfFileInput::MakeNew(const char *new_filename,
                                       FILE *new_infile,
                                       Tcl_Channel new_channel,
                                       const char *new_channelName)
{
#define MEMBERNAME "MakeNew"
  return new Vf_SvfFileInput(new_filename,new_infile,
			  new_channel,new_channelName);
#undef MEMBERNAME
}

BOOL Vf_SvfFileInput::IsType(const char * test_filename,
			  FILE *,
			  Tcl_Channel,
			  const char *test_channelName)
{
#define MEMBERNAME "IsType"
  const char* CheckProcName="Vf_SvfIsType";
  char *cmd=new char[strlen(CheckProcName)+strlen(test_filename)
	      +strlen(test_channelName)+3+16]; // 16 for safety
  sprintf(cmd,"%s %s %s",CheckProcName,test_filename,test_channelName);
  Tcl_Interp* interp=Oc_GlobalInterpreter();
  if(interp==NULL)
    FatalError(-1,STDDOC,"Tcl interpretor not initialized");
  int return_code=Tcl_Eval(interp,cmd);
  if(return_code!=TCL_OK) NonFatalError(STDDOC,"Error in Tcl proc %s: %s",
					cmd,interp->result);
  delete[] cmd;
  if(atoi(interp->result)!=0) return 1;
  return 0;
#undef MEMBERNAME
}

int Vf_SvfFileInputTA(ClientData,Tcl_Interp *interp,
		       int argc,char** argv); // Forward declaration

Vf_Mesh *Vf_SvfFileInput::NewMesh()
{
#define MEMBERNAME "NewMesh"
  // Initialize mesh, using filename as default title
  mesh=new Vf_GeneralMesh3f(filename,filename);
  mesh->ResetMesh();

  Tcl_Interp* interp=Oc_GlobalInterpreter();
  if(interp==NULL)
    FatalError(-1,STDDOC,"Tcl interpretor not initialized");

  // If needed, register Vf_SvfFileInputTA with interpreter.
  Tcl_CmdInfo info;
  Oc_AutoBuf ab;
  if(!Tcl_GetCommandInfo(interp,ab("Vf_SvfFileInputTA"),&info)) {
    // Command not registered
    Oc_RegisterCommand(interp,"Vf_SvfFileInputTA",
                       Vf_SvfFileInputTA);
  }

  // Call Tcl proc to read file and stick data into mesh
  const char* ReadProcName="Vf_SvfRead";

  char *ascii_this=new char[Omf_AsciiPtr::GetAsciiSize()];
  Omf_AsciiPtr::PtrToAscii(this,ascii_this);
  /// ascii_this is an Ascii-ified 
  int cmdlen=strlen(ReadProcName)+strlen(channelName)
		     +strlen(ascii_this)+3+16;  // 16 for safety
  char *cmd=new char[cmdlen+1];
  sprintf(cmd,"%s %s %s",ReadProcName,channelName,ascii_this);
  delete[] ascii_this;
  cmd[cmdlen]='\0';  // Safety
  int return_code=Tcl_Eval(interp,cmd);
  if(return_code!=TCL_OK) {
    NonFatalError(STDDOC,"Error in Tcl proc \"%s\": %s",
		  cmd,interp->result);
    delete[] cmd;
    delete mesh;
    return new Vf_EmptyMesh();
  }

  if(mesh->GetSize()<1) {
    ClassMessage(STDDOC,"No valid data in file %s",
		  (const char *)filename);
    delete[] cmd;
    delete mesh;
    return new Vf_EmptyMesh();
   
  }
  delete[] cmd;

  // Do post-import mesh processing
  mesh->SortPoints();

  // All done!
  return (Vf_Mesh *)mesh;
#undef MEMBERNAME
}

// Routine registered with Tcl interpreter for accessing
// Vf_SvfFileInput functions from Tcl scripts.
int Vf_SvfFileInputTA(ClientData,Tcl_Interp *interp,
		       int argc,char** argv)
{
  static char buf[SCRATCH_BUF_SIZE];
  Tcl_ResetResult(interp);
  if(argc<3) {
    sprintf(buf,"Vf_SvfFileInputTA must be called with"
	    " at least 2 arguments: <subcommand> <sfi_id> [parameters]"
	    " (%d arguments passed)",argc-1);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  Vf_SvfFileInput *sfip=(Vf_SvfFileInput *)Omf_AsciiPtr::AsciiToPtr(argv[2]);
  if(sfip==NULL || strcmp(sfip->class_doc.classname,"Vf_SvfFileInput")!=0) {
    sprintf(buf,"Sfip (%s -> %p) invalid ptr handle",argv[2],sfip);
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_ERROR;
  }

  // Switch based on sub-command parameter.  Each if-block fills
  // interp->result as appropriate, and returns TCL_OK from
  // inside the block.  The fall through case at the bottom
  // handles undefined sub-commands (and returns TCL_ERROR).

  if(strcmp(argv[1],"processline")==0) {          // Process line
    // The processline subcommand returns in interp->result a "1"
    // on a valid data line (and modifies the import string), a "0"
    // on a comment line (and leaves the import string unchanged),
    // and "-1" otherwise.  In the last case the import string will
    // most likely be altered, as in the valid data case.  At the
    // C level, TCL_OK is returned in all 3 cases.
    //  If line is a valid data line, then the input is passed into
    // the mesh, in the same manner as the addpoint subcommand.
    if(argc!=4) {
      sprintf(buf,"processline subcommand requires 2 parameters:"
	      " mesh_id line; %d passed",argc-2);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }

    // See if this is a data line or a comment
    char* line_ptr=argv[3];
    char* data_ptr=line_ptr+strspn(line_ptr," \t"); // Skip leading whitespace
    if(*data_ptr=='#') { // Comment line
      Tcl_AppendResult(interp,"0",(char *)NULL);
      return TCL_OK;
    }
    // Work under the assumption that this is a valid data line.
    REAL8m data[6];
    int i=0;
    data_ptr=strtok(data_ptr," \t");
    while(i<6 && data_ptr!=NULL && *data_ptr!='#') {
      data[i++]=Nb_Atof(data_ptr);
      data_ptr=strtok(NULL," \t");
    }
    if(i<5) {
      // Bad data line
      Tcl_AppendResult(interp,"-1",(char *)NULL);
      return TCL_OK;
    }

    // Apparently a valid line.  Copy the 5 or 6 values into a
    // LocatedVector and stick it into mesh.  These should be in the
    // order "x y z mx my mz", but some older files (version Svf-01)
    // may have only "x y mx my mz".
    static Nb_LocatedVector<REAL4> lv;
    if(i==5) {
      lv.location.Set(data[0],data[1],0.0);
      lv.value.Set(data[2],data[3],data[4]);
    }
    else {
      lv.location.Set(data[0],data[1],data[2]);
      lv.value.Set(data[3],data[4],data[5]);
    }
    sfip->mesh->AddPoint(lv);
    Tcl_AppendResult(interp,"1",(char *)NULL);
    return TCL_OK;
  }
  else if(strcmp(argv[1],"addpoint")==0) {        // Add Point
    if(argc!=9) {
      sprintf(buf,"addpoint subcommand requires 7 parameters:"
	      " mesh_id x y z mx my mz; %d passed",argc-2);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }

    static Nb_LocatedVector<REAL4> lv;
    lv.location.x=Nb_Atof(argv[3]);
    lv.location.y=Nb_Atof(argv[4]);
    lv.location.z=Nb_Atof(argv[5]);
    lv.value.x=Nb_Atof(argv[6]);
    lv.value.y=Nb_Atof(argv[7]);
    lv.value.z=Nb_Atof(argv[8]);

    // Enter data into mesh
    sfip->mesh->AddPoint(lv);
    return TCL_OK;
  }
  else if(strcmp(argv[1],"getpointcount")==0) {   // Get Point Count
    if(argc!=3) {
      sprintf(buf,"getpointcount subcommand requires 1 parameter:"
	      " mesh_id ; %d passed",argc-2);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    // Query mesh
    sprintf(buf,"%d",sfip->mesh->GetSize());
    Tcl_AppendResult(interp,buf,(char *)NULL);
    return TCL_OK;
  }
  else if(strcmp(argv[1],"setbdry")==0) {         // Set Boundary
    Nb_List< Nb_Vec3<REAL4> > boundary_list;
    if(argc>=4) { // Non-blank list
      char *pointstr=Tcl_Concat(argc-3,argv+3);
      Nb_Vec3<REAL4> v;  char* cptr;
      cptr=strtok(pointstr," \t\n\r\f");
      while(cptr!=NULL) {
	v.x=Nb_Atof(cptr);
	cptr=strtok(NULL," \t\n\r\f");
	if(cptr!=NULL) {
	  v.y=Nb_Atof(cptr);
	  v.z=0.0;
	  boundary_list.Append(v); // Add point to list
	}
	else {
	  Tcl_Free(pointstr);
	  sprintf(buf,"setbdry subcommand requires x y pairs;"
		  " %d values passed.",2*boundary_list.GetSize()+1);
	  Tcl_AppendResult(interp,buf,(char *)NULL);
	  return TCL_ERROR;
	}
	cptr=strtok(NULL," \t\n\r\f");
      }
      Tcl_Free(pointstr);
    }
    sfip->mesh->SetBoundaryList(boundary_list);
    return TCL_OK;
  }
  else if(strcmp(argv[1],"stephints")==0) {       // Grid step hints
    if(argc<4 || argc>6) {
      sprintf(buf,"addpoint subcommand requires 2-4 parameters:"
	      " mesh_id xstep [ystep [zstep]];  %d passed",argc-2);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    Nb_List< Nb_Vec3<REAL4> > step_list;
    char *pointstr=Tcl_Concat(argc-3,argv+3);
    Nb_Vec3<REAL4> step(0,0,0);  char* cptr;
    cptr=strtok(pointstr," \t\n\r\f");
    if(cptr!=NULL) {
      step.x=Nb_Atof(cptr);
      cptr=strtok(NULL," \t\n\r\f");
      if(cptr==NULL) { step.y=step.z=step.x; }
      else {
	step.y=Nb_Atof(cptr);
	cptr=strtok(NULL," \t\n\r\f");
	if(cptr!=NULL) {
	  step.z=Nb_Atof(cptr);
	}
      }
    }
    Tcl_Free(pointstr);
    sfip->mesh->SetStepHints(step);
    return TCL_OK;
  }
  else if(strcmp(argv[1],"settitle")==0) {       // Set title
    if(argc!=4) {
      sprintf(buf,"settitle subcommand requires 2 parameters:"
	      " mesh_id title;  %d passed",argc-2);
      Tcl_AppendResult(interp,buf,(char *)NULL);
      return TCL_ERROR;
    }
    sfip->mesh->SetMeshTitle(argv[3]);
    return TCL_OK;
  }

  // If we reach here, we didn't get a valid command!
  sprintf(buf,"Illegal subcommand (%s); Should be"
	  " \"processline\", \"addpoint\", \"getpointcount\","
	  " \"setbdry\",  or \"stephints\"",
	  argv[1]);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_ERROR;
}
